#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 31 04:39:06 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module contains methods relating to calculating the mutual information
between two distributions.
"""

import gaussian
import discrete
import jvhw
