#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

Created on Tue Mar  7 22:21:28 2017

@author: Daniel Speckhard, DTS@stanford.edu

This package includes exmaples to estimate the grid topology using statistical
methods. 
"""
